$RESIZE:OFF
TYPE tVECTOR2D
  x AS LONG
  y AS LONG
END TYPE

TYPE tMATERIAL
  material AS INTEGER
  behavior AS INTEGER
  flammable AS INTEGER
  pressure AS INTEGER
  sleeptime AS LONG
  update_ticks AS INTEGER
  c AS LONG ' color
  update AS INTEGER
  lifetime AS INTEGER
  n AS STRING * 20
END TYPE

TYPE tMOUSE
  position AS tVECTOR2D
  button1 AS INTEGER
  button2 AS INTEGER
  button3 AS INTEGER
  wheel AS INTEGER
END TYPE

TYPE tREACTIONS
  p AS INTEGER 'priority
  r AS INTEGER 'reaction
END TYPE

CONST SIZE_X = 200
CONST SIZE_Y = 150
CONST SCALE = 4

CONST cSLEEPTIME = 5 'iterations until cell goes to sleep

CONST cNUMOFMATERIALS = 16 ' one more than # of actual materials
CONST matNOTHING = 0
CONST matSAND = 1
CONST matLIQUID = 2
CONST matSOLID = 3
CONST matFOAM = 4 'floats on liquids
CONST matACID = 5
CONST matRESIDUE = 6
CONST matSMOKE = 7
CONST matFIRE = 8
CONST matASH = 9
CONST matACIDWATER = 10
CONST matDRAIN = 11
CONST matWATERSOURCE = 12
CONST matSTONE = 13
CONST matOIL = 14
CONST matFLAME = 15

CONST reactNOTHING = 0
CONST reactAIR = 0
CONST reactSAND = 1
CONST reactWATER = 2
CONST reactWOOD = 3
CONST reactFOAM = 4 'floats on liquids
CONST reactACID = 5
CONST reactRESIDUE = 6
CONST reactSMOKE = 7
CONST reactFIRE = 8
CONST reactASH = 9
CONST reactACIDWATER = 10
CONST reactDRAIN = 11
CONST reactWATERSOURCE = 12
CONST reactSTONE = 13
CONST reactOIL = 14
CONST reactFLAME = 15

DIM movement_matrix(cNUMOFMATERIALS, cNUMOFMATERIALS, 8) AS INTEGER



_TITLE "Cellular Automata"
SCREEN _NEWIMAGE(SIZE_X * SCALE + 200, SIZE_Y * SCALE, 32)
DIM SHARED AS LONG hsc
hsc = _NEWIMAGE(_WIDTH, _HEIGHT, 32)
_PRINTMODE _KEEPBACKGROUND
DIM cell(SIZE_X, SIZE_Y) AS tMATERIAL

DIM mouse AS tMOUSE
DIM material_prop(20) AS tMATERIAL
DIM SHARED freeze, quit AS INTEGER
screenmiddle

' init_movements(movement_matrix())
openCSV "./move1.csv", movement_matrix()
materialInitialize material_prop()

'*** Main Loop
DO
  CLS , _RGB32(180)

  'If Not freeze Then Call run_sim(cell(), material_prop())
  IF NOT freeze THEN runSimEx cell(), material_prop(), movement_matrix()
  render cell()
  renderMenu material_prop(), mouse
  handleKeypress cell(), material_prop()
  handleMouse cell(), material_prop(), mouse
  _LIMIT 240
  _DISPLAY


LOOP UNTIL quit = 1
SYSTEM

'***  Subs and functions

SUB renderMenu (mat_prop() AS tMATERIAL, mouse AS tMOUSE)
  DIM i AS INTEGER
  DIM sx AS INTEGER
  DIM sy AS INTEGER
  sx = SIZE_X * SCALE
  sy = SIZE_Y * SCALE
  fontwidth = 8
  fontheight = 16
  LINE (0, 0)-(sx - 1, sy - 1), _RGB(50, 67, 127), B

  FOR i = 0 TO cNUMOFMATERIALS - 1
    ' Locate i + 1 * 10, 104
    LOCATE i + 1 * 10, (sx / fontwidth) + 4
    PRINT mat_prop(i).n
    IF i = ABS(mouse.wheel) THEN
      LINE (102 * fontwidth, i * fontheight + 143)-(116 * fontwidth, i * fontheight + 158), , B
    END IF
    _DEST hsc
    LINE (102 * fontwidth, i * fontheight + 143)-(116 * fontwidth, i * fontheight + 158), _RGB32(i + 5, 0, 0), BF
    _DEST 0

  NEXT

END SUB

SUB handleKeypress (cell() AS tMATERIAL, mat_prop() AS tMATERIAL)
  DIM keys AS LONG
  keys = _KEYHIT
  IF keys = 27 THEN quit = 1
  IF keys = 32 THEN freeze = NOT (freeze)

  IF keys > 0 AND keys < 256 THEN
    IF LCASE$(CHR$(keys)) = "c" THEN
      clearCells cell(), mat_prop()
    END IF
    IF LCASE$(CHR$(keys)) = "s" THEN
      saveCells cell(), "CellAuto.ca"
    END IF
    IF LCASE$(CHR$(keys)) = "l" THEN
      loadCells cell(), "CellAuto.ca"
    END IF
  END IF
END SUB

SUB saveCells (cell() AS tMATERIAL, filename AS STRING)
  DIM i, j AS INTEGER

  OPEN filename FOR OUTPUT AS #1
  PRINT #1, SIZE_X
  PRINT #1, SIZE_Y
  FOR j = 0 TO SIZE_Y
    FOR i = 0 TO SIZE_X
      PRINT #1, cell(i, j).material
      PRINT #1, cell(i, j).behavior
      PRINT #1, cell(i, j).flammable
      PRINT #1, cell(i, j).sleeptime
      PRINT #1, cell(i, j).update_ticks
      PRINT #1, cell(i, j).c
      PRINT #1, cell(i, j).update
      PRINT #1, cell(i, j).lifetime
      PRINT #1, cell(i, j).n
    NEXT
  NEXT
  CLOSE #1
END SUB

SUB loadCells (cell() AS tMATERIAL, filename AS STRING)
  DIM i, j, Sx, Sy AS INTEGER

  OPEN filename FOR INPUT AS #1
  INPUT #1, Sx
  INPUT #1, Sy
  FOR j = 0 TO Sy
    FOR i = 0 TO Sx
      INPUT #1, cell(i, j).material
      INPUT #1, cell(i, j).behavior
      INPUT #1, cell(i, j).flammable
      INPUT #1, cell(i, j).sleeptime
      INPUT #1, cell(i, j).update_ticks
      INPUT #1, cell(i, j).c
      INPUT #1, cell(i, j).update
      INPUT #1, cell(i, j).lifetime
      INPUT #1, cell(i, j).n
    NEXT
  NEXT
  CLOSE #1
END SUB


SUB clearCells (cell() AS tMATERIAL, mat_prop() AS tMATERIAL)
  DIM j, i AS INTEGER
  FOR j = 0 TO SIZE_Y
    FOR i = 0 TO SIZE_X
      cell(i, j) = mat_prop(matNOTHING)
    NEXT
  NEXT
END SUB

SUB runSimEx (cell() AS tMATERIAL, mat_prop() AS tMATERIAL, mm() AS INTEGER)
  DIM i, j, k, p, xD, yD, w, wc, m, rng AS INTEGER
  DIM AS INTEGER ii, jj
  DIM movement_priority_list(9, 9) AS INTEGER
  DIM movement_priority_count(9) AS INTEGER
  FOR j = 1 TO SIZE_Y - 1
    FOR i = 1 TO SIZE_X - 1
      IF cell(i, j).update = 0 THEN ' update tick are used to simulate viscosity. *cant spell
        IF cell(i, j).sleeptime > 0 THEN
          w = cell(i, j).material
          FOR k = 0 TO 8

            xD = (k MOD 3) - 1 ' calculate perimeter offsets
            yD = INT(k / 3) - 1


            wc = cell(i + xD, j + yD).material
            IF k <> 4 THEN
              'Handle reactions beween materials
              SELECT CASE w
                CASE reactAIR
                  IF k > 5 AND wc = reactFIRE THEN
                    setMaterial cell(), mat_prop(reactFLAME), i, j
                  END IF
                CASE reactFLAME
                  cell(i, j).sleeptime = 10
                  cell(i, j).update = 0
                  IF checkPresence(cell(), reactWATER, i, j) > -1 THEN
                    setMaterial cell(), mat_prop(reactWATER), i, j
                  END IF
                  IF checkPresence(cell(), reactFIRE, i, j) = -1 THEN
                    cell(i, j).lifetime = cell(i, j).lifetime - 1
                    IF cell(i, j).lifetime < 1 THEN
                      setMaterial cell(), mat_prop(reactAIR), i, j
                    END IF
                  END IF

                  SELECT CASE wc
                    CASE reactWOOD:
                      setMaterial cell(), mat_prop(reactFIRE), i, j
                      setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                      IF RND > .5 THEN
                        cell(i, j).c = _RGB32(205, 177, 6)
                      ELSE
                        cell(i, j).c = _RGB32(211, 50, 0)
                      END IF

                    CASE reactFOAM:
                      setMaterial cell(), mat_prop(reactFIRE), i, j
                      setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                      IF RND > .5 THEN
                        cell(i, j).c = _RGB32(205, 177, 6)
                      ELSE
                        cell(i, j).c = _RGB32(211, 50, 0)
                      END IF

                    CASE reactOIL:
                      setMaterial cell(), mat_prop(reactFIRE), i, j
                      setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                      IF RND > .5 THEN
                        cell(i, j).c = _RGB32(205, 177, 6)
                      ELSE
                        cell(i, j).c = _RGB32(211, 50, 0)
                      END IF

                    CASE ELSE
                      cell(i, j).lifetime = cell(i, j).lifetime - 1
                      IF cell(i, j).lifetime < 1 THEN
                        IF RND > .5 THEN
                          setMaterial cell(), mat_prop(matFLAME), i, j
                        ELSE
                          setMaterial cell(), mat_prop(matASH), i, j
                        END IF
                      END IF
                  END SELECT
                  cell(i, j).lifetime = cell(i, j).lifetime - 1
                  IF cell(i, j).lifetime < 1 THEN
                    IF RND > .75 THEN
                      setMaterial cell(), mat_prop(reactSMOKE), i, j
                    ELSE
                      setMaterial cell(), mat_prop(reactAIR), i, j
                    END IF
                  END IF


                CASE reactACID:
                  SELECT CASE wc
                    CASE reactWOOD:
                      setMaterial cell(), mat_prop(reactRESIDUE), i + xD, j + yD
                    CASE reactFOAM:
                      setMaterial cell(), mat_prop(reactRESIDUE), i, j
                      setMaterial cell(), mat_prop(reactSMOKE), i + xD, j + yD
                    CASE reactWATER:
                      setMaterial cell(), mat_prop(reactACIDWATER), i, j
                      setMaterial cell(), mat_prop(reactACIDWATER), i + xD, j + yD
                  END SELECT
                CASE reactFIRE:
                  'Fire will extinguish without air or in the presence of water
                  cell(i, j).sleeptime = 10
                  IF checkPresence(cell(), reactWATER, i, j) > -1 THEN
                    setMaterial cell(), mat_prop(reactASH), i, j
                  END IF

                  IF checkPresence(cell(), reactAIR, i, j) = -1 OR checkPresence(cell(), reactWATER, i, j) > -1 OR checkPresence(cell(), reactFLAME, i, j) = -1 THEN
                    cell(i, j).lifetime = cell(i, j).lifetime - 1
                    IF cell(i, j).lifetime < 1 THEN
                      IF RND > .5 THEN
                        setMaterial cell(), mat_prop(reactFLAME), i, j
                      ELSE
                        setMaterial cell(), mat_prop(reactASH), i, j
                      END IF
                    END IF
                  ELSE
                    SELECT CASE wc
                      CASE reactWOOD:
                        setMaterial cell(), mat_prop(reactFIRE), i, j
                        setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                        IF RND > .5 THEN
                          cell(i, j).c = _RGB32(205, 177, 6)
                        ELSE
                          cell(i, j).c = _RGB32(211, 50, 0)
                        END IF

                      CASE reactFOAM:
                        setMaterial cell(), mat_prop(reactFIRE), i, j
                        setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                        IF RND > .5 THEN
                          cell(i, j).c = _RGB32(205, 177, 6)
                        ELSE
                          cell(i, j).c = _RGB32(211, 50, 0)
                        END IF

                      CASE reactOIL:
                        setMaterial cell(), mat_prop(reactFIRE), i, j
                        setMaterial cell(), mat_prop(reactFIRE), i + xD, j + yD
                        IF RND > .5 THEN
                          cell(i, j).c = _RGB32(205, 177, 6)
                        ELSE
                          cell(i, j).c = _RGB32(211, 50, 0)
                        END IF

                      CASE ELSE
                        cell(i, j).lifetime = cell(i, j).lifetime - 1
                        IF cell(i, j).lifetime < 1 THEN
                          IF RND > .5 THEN
                            setMaterial cell(), mat_prop(matFLAME), i, j
                          ELSE
                            setMaterial cell(), mat_prop(matASH), i, j
                          END IF
                        END IF
                    END SELECT
                  END IF
                CASE reactACIDWATER:
                  SELECT CASE wc
                    CASE reactWATER:
                      IF k > 5 THEN ' only mix with whats below
                        setMaterial cell(), mat_prop(reactACIDWATER), i + xD, j + yD
                      END IF
                  END SELECT
                  ' update if material changes, so movement matrix will be correct
                  w = cell(i, j).material
                  wc = cell(i + xD, j + yD).material
                CASE reactWATERSOURCE:
                  FOR jj = -1 TO 1
                    FOR ii = -1 TO 1
                      IF jj <> ii THEN
                        setMaterial cell(), mat_prop(matLIQUID), ii + i, jj + j
                      END IF
                    NEXT
                  NEXT
                CASE reactDRAIN:
                  FOR jj = -1 TO 1
                    FOR ii = -1 TO 1
                      IF jj <> ii THEN
                        setMaterial cell(), mat_prop(matNOTHING), ii + i, jj + j
                      END IF
                    NEXT
                  NEXT

              END SELECT


              m = mm(w, wc, k) ' movement matrix comparison between two materials
              IF m > 0 THEN
                movement_priority_list(m, movement_priority_count(m)) = k
                movement_priority_count(m) = movement_priority_count(m) + 1
              END IF
            END IF
          NEXT
          'Handle the movement of cells after the reactions are complete
          'Start at the high priority and move down until a movement can be made
          FOR p = 1 TO 8 ' start at 1 because 0 is doing nothing
            IF movement_priority_count(p) > 0 THEN 'is there anything in the list?
              rng = INT(RND * movement_priority_count(p)) 'randomize if more than one option
              xD = (movement_priority_list(p, rng) MOD 3) - 1
              yD = INT(movement_priority_list(p, rng) / 3) - 1
              moveMaterial cell(), mat_prop(), i, j, i + xD, j + yD
              EXIT FOR
            END IF
          NEXT
          'clear for the next go round
          FOR p = 0 TO 9
            movement_priority_count(p) = 0
          NEXT
          cell(i, j).sleeptime = cell(i, j).sleeptime - 1
        END IF
      END IF
    NEXT
  NEXT
END SUB

FUNCTION checkPresence (cell() AS tMATERIAL, mat AS INTEGER, x AS INTEGER, y AS INTEGER)
  DIM p, k, xD, yD AS INTEGER
  p = -1
  FOR k = 0 TO 8
    xD = ((k MOD 3) - 1) + x ' calculate perimeter offsets
    yD = (INT(k / 3) - 1) + y
    IF xD >= 0 AND xD <= SIZE_X AND yD >= 0 AND yD <= SIZE_Y THEN
      IF cell(xD, yD).material = mat THEN
        p = k
        EXIT FOR
      END IF
    END IF
  NEXT
  checkPresence = p
END FUNCTION


SUB render (cell() AS tMATERIAL)
  DIM i, j AS INTEGER
  FOR j = 1 TO SIZE_Y - 1
    FOR i = 1 TO SIZE_X - 1
      LINE (i * SCALE, j * SCALE)-(i * SCALE + (SCALE - 1), j * SCALE + (SCALE - 1)), cell(i, j).c, BF
      IF cell(i, j).update > 0 THEN
        cell(i, j).update = cell(i, j).update - 1 ' decrement tick count for next update
      END IF
    NEXT
  NEXT
END SUB

'SUB setMaterial (cell() AS tMATERIAL, mat_prop() AS tMATERIAL, material AS INTEGER, x AS INTEGER, y AS INTEGER)
'  cell(x, y) = mat_prop(material)
'  updateCellLifetime cell(), x, y, cSLEEPTIME ' wake up neighboring cells
'END SUB


SUB setMaterial (cell() AS tMATERIAL, mat_prop AS tMATERIAL, x AS INTEGER, y AS INTEGER)
  cell(x, y) = mat_prop
  updateCellLifetime cell(), x, y, cSLEEPTIME ' wake up neighboring cells
END SUB


SUB moveMaterial (cell() AS tMATERIAL, mat_prop() AS tMATERIAL, x AS INTEGER, y AS INTEGER, xm AS INTEGER, ym AS INTEGER)
  SWAP cell(x, y), cell(xm, ym)
  cell(x, y).update = mat_prop(cell(x, y).material).update_ticks ' cell has been processed this cycle
  cell(xm, ym).update = mat_prop(cell(xm, ym).material).update_ticks ' cell has been processed this cycle
  updateCellLifetime cell(), x, y, cSLEEPTIME ' wake up neighboring cells
  updateCellLifetime cell(), xm, ym, cSLEEPTIME ' wake up neighboring cells
END SUB

SUB updateCellLifetime (cell() AS tMATERIAL, x AS INTEGER, y AS INTEGER, l AS INTEGER)
  'wake up neighboring cells as they may be affected by nearby changes
  FOR j = -1 TO 1
    FOR i = -1 TO 1
      IF x + i > 0 AND y + j > 0 AND x + i < SIZE_X AND y + j < SIZE_Y THEN
        ' IF cell(i + x, j + y).material > matNOTHING THEN
        cell(x + i, y + j).sleeptime = l
        ' END IF
      END IF
    NEXT
  NEXT
END SUB


SUB handleMouse (cell() AS tMATERIAL, mat_prop() AS tMATERIAL, mouse AS tMOUSE)
  DIM w, i, j AS INTEGER
  DIM AS LONG hscc
  DO WHILE _MOUSEINPUT
    mouse.position.x = _MOUSEX: mouse.position.y = _MOUSEY
    mouse.button1 = _MOUSEBUTTON(1)
    mouse.button2 = _MOUSEBUTTON(2)
    mouse.button3 = _MOUSEBUTTON(3)
    mouse.wheel = ((mouse.wheel + _MOUSEWHEEL) MOD cNUMOFMATERIALS)
  LOOP
  w = ABS(mouse.wheel)
  IF mouse.button1 THEN
    _SOURCE hsc
    hscc = _RED32(POINT(mouse.position.x, mouse.position.y))
    IF hscc > 0 THEN mouse.wheel = hscc - 5
    _SOURCE 0

    IF mouse.position.x > 1 AND mouse.position.x < SIZE_X * SCALE - 1 AND mouse.position.y > 1 AND mouse.position.y < SIZE_Y * SCALE - 1 THEN
      FOR j = -1 TO 1
        FOR i = -1 TO 1
          setMaterial cell(), mat_prop(w), mouse.position.x / SCALE + i, mouse.position.y / SCALE + j
        NEXT
      NEXT
    ELSE
    END IF
  END IF
END SUB


SUB openCSV (f AS STRING, m() AS INTEGER)
  DIM ln AS STRING
  DIM d AS STRING
  DIM AS INTEGER x, y

  OPEN f FOR INPUT AS #1
  LINE INPUT #1, ln ' discard first line
  y = 0
  DO UNTIL EOF(1)
    LINE INPUT #1, ln
    cp = INSTR(ln, ":")
    x = 0
    DO UNTIL cp = 0
      d = MID$(ln, cp + 1, 9)
      FOR i = 1 TO LEN(d)
        m(x, y, i - 1) = VAL(MID$(d, i, 1))
      NEXT i
      cp = INSTR(cp + 1, ln, ":")
      x = x + 1
    LOOP
    y = y + 1
  LOOP
  CLOSE #1
END SUB


SUB materialInitialize (ms() AS tMATERIAL)
  DIM i, r, g, b, behavior, update, life AS INTEGER

  DATA "air",0,0,180,180,180,-1
  DATA "sand",1,1,166,172,127,-1
  DATA "water",2,1,67,72,183,-1
  DATA "wood",3,1,111,72,0,-1
  DATA "foam",4,1,255,255,255,-1
  DATA "acid",2,4,17,216,22,-1
  DATA "residue",1,6,127,67,127,-1
  DATA "smoke",7,10,128,127,127,50
  DATA "fire",1,20,255,64,0,2000
  DATA "ash",1,4,32,32,32,-1
  DATA "acidwater",2,10,6,83,6,-1
  DATA "drain",11,4,44,28,28,-1
  DATA "water source",12,1,44,28,28,-1
  DATA "stone",13,1,127,127,127,-1
  DATA "oil",14,2,20,20,20,-1
  DATA "flame",15,5,255,64,0,50


  FOR i = 0 TO cNUMOFMATERIALS - 1
    READ ms(i).n
    READ behavior
    READ update
    READ r
    READ g
    READ b
    READ life
    ms(i).c = _RGB32(r, g, b)
    ms(i).behavior = behavior ' currently unused
    ms(i).material = i
    ms(i).sleeptime = cSLEEPTIME
    ms(i).update = update
    ms(i).update_ticks = update
    ms(i).lifetime = life
  NEXT i
END SUB

SUB screenmiddle () ' by mnrvovrfc
  DIM AS INTEGER dw, dh, ww, hh
  dw = _DESKTOPWIDTH
  dh = _DESKTOPHEIGHT
  ww = _WIDTH
  hh = _HEIGHT
  ww = (dw \ 2) - (ww \ 2)
  hh = (dh \ 2) - (hh \ 2)
  _SCREENMOVE ww - 500, hh - 300
END SUB

